import numpy as np
from tqdm import tqdm
from typing import Optional, List, Tuple
from .helper import check_shape, ensure_float32
from .basic import window_slice, jitter
from .dtw import DTW

# Example docstring for a label-dependent augmentation:
def spawner(x, labels, sigma=0.05, verbose=0, random_state=None) -> Tuple[np.ndarray, List[int]]:
    """
    SPAWNER augmentation: averages two intra-class patterns using DTW path.
    Returns:
        - Augmented data (np.ndarray)
        - List of indices where no augmentation was possible (skipped indices)
    Reference: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6983028/
    """
    check_shape(x)
    rng = np.random.RandomState(random_state)
    random_points = rng.randint(low=1, high=x.shape[1]-1, size=x.shape[0])
    window = np.ceil(x.shape[1] / 10.).astype(int)
    orig_steps = np.arange(x.shape[1])
    l = np.argmax(labels, axis=1) if labels.ndim > 1 else labels
    ret = np.zeros_like(x)
    skipped = []
    for i, pat in enumerate(tqdm(x)):
        choices = np.delete(np.arange(x.shape[0]), i)
        choices = np.where(l[choices] == l[i])[0]
        if choices.size > 0:
            random_sample = x[rng.choice(choices)]
            path1 = DTW.dtw(pat[:random_points[i]], random_sample[:random_points[i]], DTW.RETURN_PATH, slope_constraint="symmetric", window=window)
            path2 = DTW.dtw(pat[random_points[i]:], random_sample[random_points[i]:], DTW.RETURN_PATH, slope_constraint="symmetric", window=window)
            combined = np.concatenate((np.vstack(path1), np.vstack(path2+random_points[i])), axis=1)
            mean = np.mean([pat[combined[0]], random_sample[combined[1]]], axis=0)
            for dim in range(x.shape[2]):
                ret[i,:,dim] = np.interp(orig_steps, np.linspace(0, x.shape[1]-1., num=mean.shape[0]), mean[:,dim]).T
        else:
            if verbose > -1:
                print(f"Only one pattern of class {l[i]}, skipping pattern average.")
            ret[i,:] = pat
            skipped.append(i)
    return ensure_float32(jitter(ret, sigma=sigma, random_state=random_state)), skipped

def wdba(
    x: np.ndarray, 
    labels: np.ndarray, 
    batch_size: int = 6, 
    slope_constraint: str = "symmetric", 
    use_window: bool = True, 
    verbose: int = 0, 
    random_state: Optional[int] = None
) -> Tuple[np.ndarray, List[int]]:
    """
    Weighted DBA (DTW Barycenter Averaging) augmentation.
    Returns augmented data and list of skipped indices.
    Reference: https://ieeexplore.ieee.org/document/8215569
    """
    check_shape(x)
    rng = np.random.RandomState(random_state)
    window = np.ceil(x.shape[1] / 10.).astype(int) if use_window else None
    orig_steps = np.arange(x.shape[1])
    l = np.argmax(labels, axis=1) if labels.ndim > 1 else labels
    ret = np.zeros_like(x)
    skipped = []
    for i in tqdm(range(ret.shape[0])):
        choices = np.where(l == l[i])[0]
        if choices.size > 0:
            k = min(choices.size, batch_size)
            random_prototypes = x[rng.choice(choices, k, replace=False)]
            # DTW matrix
            dtw_matrix = np.zeros((k, k))
            for p, prototype in enumerate(random_prototypes):
                for s, sample in enumerate(random_prototypes):
                    if p == s:
                        dtw_matrix[p, s] = 0.
                    else:
                        dtw_matrix[p, s] = DTW.dtw(
                            prototype, sample, DTW.RETURN_VALUE, 
                            slope_constraint=slope_constraint, window=window
                        )
            medoid_id = np.argsort(np.sum(dtw_matrix, axis=1))[0]
            nearest_order = np.argsort(dtw_matrix[medoid_id])
            medoid_pattern = random_prototypes[medoid_id]
            average_pattern = np.zeros_like(medoid_pattern)
            weighted_sums = np.zeros((medoid_pattern.shape[0]))
            for nid in nearest_order:
                if nid == medoid_id or dtw_matrix[medoid_id, nearest_order[1]] == 0.:
                    average_pattern += medoid_pattern
                    weighted_sums += np.ones_like(weighted_sums)
                else:
                    idx1, idx2 = DTW.dtw(
                        medoid_pattern, random_prototypes[nid], DTW.RETURN_PATH,
                        slope_constraint=slope_constraint, window=window
                    )
                    dtw_value = dtw_matrix[medoid_id, nid]
                    warped = random_prototypes[nid][idx2]
                    weight = np.exp(np.log(0.5) * dtw_value / (dtw_matrix[medoid_id, nearest_order[1]] + 1e-8))
                    average_pattern[idx1] += weight * warped
                    weighted_sums[idx1] += weight
            ret[i, :] = average_pattern / (weighted_sums[:, np.newaxis] + 1e-8)
        else:
            if verbose > -1:
                print(f"Only one pattern of class {l[i]}, skipping pattern average.")
            ret[i, :] = x[i]
            skipped.append(i)
    return ensure_float32(ret), skipped

def random_guided_warp(
    x: np.ndarray, 
    labels: np.ndarray, 
    slope_constraint: str = "symmetric", 
    use_window: bool = True, 
    dtw_type: str = "normal", 
    verbose: int = 0, 
    random_state: Optional[int] = None
) -> Tuple[np.ndarray, List[int]]:
    """
    Random guided DTW warping using intra-class prototypes.
    Returns augmented data and list of skipped indices.
    """
    check_shape(x)
    rng = np.random.RandomState(random_state)
    window = np.ceil(x.shape[1] / 10.).astype(int) if use_window else None
    orig_steps = np.arange(x.shape[1])
    l = np.argmax(labels, axis=1) if labels.ndim > 1 else labels
    ret = np.zeros_like(x)
    skipped = []
    for i, pat in enumerate(tqdm(x)):
        choices = np.delete(np.arange(x.shape[0]), i)
        choices = np.where(l[choices] == l[i])[0]
        if choices.size > 0:
            random_prototype = x[rng.choice(choices)]
            if dtw_type == "shape":
                raise NotImplementedError("shapeDTW is not implemented.")
            else:
                idx1, idx2 = DTW.dtw(random_prototype, pat, DTW.RETURN_PATH, slope_constraint=slope_constraint, window=window)
            warped = pat[idx2]
            for dim in range(x.shape[2]):
                ret[i, :, dim] = np.interp(orig_steps, np.linspace(0, x.shape[1]-1., num=warped.shape[0]), warped[:, dim]).T
        else:
            if verbose > -1:
                print(f"Only one pattern of class {l[i]}, skipping timewarping.")
            ret[i, :] = pat
            skipped.append(i)
    return ensure_float32(ret), skipped

def discriminative_guided_warp(
    x: np.ndarray, 
    labels: np.ndarray, 
    batch_size: int = 6, 
    slope_constraint: str = "symmetric", 
    use_window: bool = True, 
    dtw_type: str = "normal", 
    use_variable_slice: bool = True, 
    verbose: int = 0, 
    random_state: Optional[int] = None
) -> Tuple[np.ndarray, List[int]]:
    """
    Discriminative guided warping based on intra- and inter-class prototypes.
    Returns augmented data and list of skipped indices.
    """
    check_shape(x)
    rng = np.random.RandomState(random_state)
    window = np.ceil(x.shape[1] / 10.).astype(int) if use_window else None
    orig_steps = np.arange(x.shape[1])
    l = np.argmax(labels, axis=1) if labels.ndim > 1 else labels
    positive_batch = int(np.ceil(batch_size / 2))
    negative_batch = int(np.floor(batch_size / 2))
    ret = np.zeros_like(x)
    warp_amount = np.zeros(x.shape[0])
    skipped = []
    for i, pat in enumerate(tqdm(x)):
        choices = np.delete(np.arange(x.shape[0]), i)
        positive = np.where(l[choices] == l[i])[0]
        negative = np.where(l[choices] != l[i])[0]
        if positive.size > 0 and negative.size > 0:
            pos_k = min(positive.size, positive_batch)
            neg_k = min(negative.size, negative_batch)
            positive_prototypes = x[rng.choice(positive, pos_k, replace=False)]
            negative_prototypes = x[rng.choice(negative, neg_k, replace=False)]
            pos_aves = np.zeros((pos_k))
            neg_aves = np.zeros((pos_k))
            if dtw_type == "shape":
                raise NotImplementedError("shapeDTW is not implemented.")
            else:
                for p, pos_prot in enumerate(positive_prototypes):
                    for ps, pos_samp in enumerate(positive_prototypes):
                        if p != ps:
                            pos_aves[p] += (1./(pos_k-1.))*DTW.dtw(pos_prot, pos_samp, DTW.RETURN_VALUE, slope_constraint=slope_constraint, window=window)
                    for ns, neg_samp in enumerate(negative_prototypes):
                        neg_aves[p] += (1./neg_k)*DTW.dtw(pos_prot, neg_samp, DTW.RETURN_VALUE, slope_constraint=slope_constraint, window=window)
                selected_id = np.argmax(neg_aves - pos_aves)
                idx1, idx2 = DTW.dtw(positive_prototypes[selected_id], pat, DTW.RETURN_PATH, slope_constraint=slope_constraint, window=window)
            warped = pat[idx2]
            warp_path_interp = np.interp(orig_steps, np.linspace(0, x.shape[1]-1., num=warped.shape[0]), idx2)
            warp_amount[i] = np.sum(np.abs(orig_steps-warp_path_interp))
            for dim in range(x.shape[2]):
                ret[i, :, dim] = np.interp(orig_steps, np.linspace(0, x.shape[1]-1., num=warped.shape[0]), warped[:, dim]).T
        else:
            if verbose > -1:
                print(f"Only one pattern of class {l[i]}, skipping discriminative warping.")
            ret[i, :] = pat
            warp_amount[i] = 0.
            skipped.append(i)
    if use_variable_slice:
        max_warp = np.max(warp_amount)
        if max_warp == 0:
            ret = window_slice(ret, reduce_ratio=0.9)
        else:
            for i, pat in enumerate(ret):
                ret[i] = window_slice(pat[np.newaxis, :, :], reduce_ratio=0.9+0.1*warp_amount[i]/max_warp)[0]
    return ensure_float32(ret), skipped


from tslearn.metrics import dtw_path, dtw
import numpy as np

class DTW:
    RETURN_PATH = "path"
    RETURN_VALUE = "value"
    RETURN_ALL = "all"
    
    @staticmethod
    def dtw(ts1, ts2, return_flag="path", slope_constraint="symmetric", window=None):
        """
        ts1, ts2: np.ndarray, shape [seq_len, n_features]
        return_flag: 'path', 'value', or 'all'
        slope_constraint: unused for tslearn (tslearn always symmetric)
        window: ignored for now (can use Sakoe-Chiba band if needed)
        """
        # tslearn expects 2D arrays
        if ts1.ndim == 1:
            ts1 = ts1[:, None]
        if ts2.ndim == 1:
            ts2 = ts2[:, None]
        # Optionally handle windowing (Sakoe-Chiba band)
        if window is not None:
            # tslearn uses sakoe_chiba_radius, which is half the window size
            sakoe_chiba_radius = int(window // 2)
        else:
            sakoe_chiba_radius = None
        path, dist = dtw_path(ts1, ts2, global_constraint="sakoe_chiba" if sakoe_chiba_radius else None, sakoe_chiba_radius=sakoe_chiba_radius)
        if return_flag == DTW.RETURN_PATH:
            # Unzip path to index arrays
            idx1, idx2 = zip(*path)
            return np.array(idx1), np.array(idx2)
        elif return_flag == DTW.RETURN_VALUE:
            return dist
        elif return_flag == DTW.RETURN_ALL:
            idx1, idx2 = zip(*path)
            return dist, None, None, (np.array(idx1), np.array(idx2))
        else:
            raise ValueError(f"Unknown return_flag: {return_flag}")
        

    @staticmethod
    def shapeDTW(ts1, ts2, window_size=5, dist_func=None, return_flag="path"):
        """
        shapeDTW: Dynamic Time Warping with local shape descriptors.
        
        Args:
            ts1, ts2: np.ndarray, shape (seq_len, n_features)
            window_size: int, size of window for local descriptors (must be odd, default 5)
            dist_func: callable, function to compute distance between two windows (default: Euclidean)
            return_flag: "path" (indices), "value" (total cost), or "all"
        Returns:
            Depending on return_flag:
            - "path": idx1, idx2 (warping path indices)
            - "value": total cost (float)
            - "all": total cost, None, None, (idx1, idx2)
        """
        assert ts1.ndim == 2 and ts2.ndim == 2, "shapeDTW: ts1 and ts2 must be 2D arrays"
        assert window_size % 2 == 1, "shapeDTW: window_size should be odd"
        if dist_func is None:
            def dist_func(w1, w2):
                return np.linalg.norm(w1 - w2)

        half = window_size // 2
        # Pad the series to handle borders
        ts1_padded = np.pad(ts1, ((half, half), (0, 0)), mode='reflect')
        ts2_padded = np.pad(ts2, ((half, half), (0, 0)), mode='reflect')
        # Extract all windows
        descs1 = np.array([ts1_padded[i:i+window_size] for i in range(ts1.shape[0])])
        descs2 = np.array([ts2_padded[i:i+window_size] for i in range(ts2.shape[0])])
        # Compute cost matrix
        cost_matrix = np.zeros((ts1.shape[0], ts2.shape[0]))
        for i in range(ts1.shape[0]):
            for j in range(ts2.shape[0]):
                cost_matrix[i, j] = dist_func(descs1[i], descs2[j])
        # Use tslearn DTW on cost matrix (use 1D cost matrix as multivariate)
        # We'll use dtw_path on cost matrix directly
        path, cost = dtw_path(cost_matrix.reshape(-1, 1), np.zeros((cost_matrix.shape[1], 1)))
        idx1, idx2 = zip(*path)
        if return_flag == "path":
            return np.array(idx1), np.array(idx2)
        elif return_flag == "value":
            return cost
        elif return_flag == "all":
            return cost, None, None, (np.array(idx1), np.array(idx2))
        else:
            raise ValueError(f"Unknown return_flag: {return_flag}")


import torch

EPS = 1e-8

Z_VALUE_CAP = 5.0 #(random) suggestion by an LLM 

def stablize(z):
    z = z + EPS * torch.randn_like(z)
    z = torch.nan_to_num(z, nan=0.0, posinf=Z_VALUE_CAP, neginf=-Z_VALUE_CAP)
    z = torch.clamp(z, -Z_VALUE_CAP, Z_VALUE_CAP)
    return z


import mlflow
import mlflow.pytorch
import torch
import torch.optim as optim
from torch.utils.data import DataLoader
from tqdm import tqdm
import logging
from typing import Dict
from torchdtcc.dtcc.dtcc import DTCC
from torchdtcc.dtcc.clustering import Clusterer
from torchdtcc.datasets.augmented_dataset import AugmentedDataset
from .trainer import DTCCTrainer

class MlFlowDTCCTrainer(DTCCTrainer):
    def __init__(
        self,
        model: DTCC,
        dataloader: DataLoader,
        augment_time_series,
        optimizer,
        lambda_cd,
        num_epochs,
        update_interval=5,
        device="cpu",
        experiment_name: str = "MLflow_DTCC_Training",
        run_name: str = "default_run"
    ):
        super().__init__(model, dataloader, augment_time_series, optimizer, lambda_cd, num_epochs, update_interval, device)
        self.experiment_name = experiment_name
        self.run_name = run_name
        mlflow.set_experiment(experiment_name)
        logging.info(f"MLflow experiment set to: {experiment_name}")

    def run(self, save_path=None):
        with mlflow.start_run(run_name=self.run_name):
            mlflow.log_param("lambda_cd", self.lambda_cd)
            mlflow.log_param("num_epochs", self.num_epochs)
            mlflow.log_param("update_interval", self.update_interval)
            if hasattr(self.optimizer, 'param_groups'):
                mlflow.log_param("learning_rate", self.optimizer.param_groups[0]['lr'])
                mlflow.log_param("weight_decay", self.optimizer.param_groups[0].get('weight_decay', 0))
            result = super().run(save_path)
        return result
    
    def log_loss(self, epoch, avg_recon, avg_instance, avg_cd, avg_cluster, avg_total):
        # Log epoch metrics to MLflow
        mlflow.log_metric("avg_recon_loss", avg_recon, step=epoch + 1)
        mlflow.log_metric("avg_instance_loss", avg_instance, step=epoch + 1)
        mlflow.log_metric("avg_cd_loss", avg_cd, step=epoch + 1)
        mlflow.log_metric("avg_cluster_loss", avg_cluster, step=epoch + 1)
        mlflow.log_metric("avg_total_loss", avg_total, step=epoch + 1)

        logging.info(
            f"Epoch {epoch+1}/{self.num_epochs} | avg recon: {avg_recon:.4f} | avg instance: {avg_instance:.4f} | avg cd: {avg_cd:.4f} | avg cluster: {avg_cluster:.4f} | avg total: {avg_total:.4f}"
        )
    
    def log_evaluation(self, epoch, metrics):
        # Log clustering metrics to MLflow
        mlflow.log_metric("ACC", metrics['acc'], step=epoch + 1)
        mlflow.log_metric("NMI", metrics['nmi'], step=epoch + 1)
        mlflow.log_metric("ARI", metrics['ari'], step=epoch + 1)
        mlflow.log_metric("RI", metrics['ri'], step=epoch + 1)
        print(f"Epoch {epoch+1}: ACC={metrics['acc']:.4f} NMI={metrics['nmi']:.4f} ARI={metrics['ari']:.4f} RI={metrics['ri']:.4f}")

    def save_model(self, save_path):
        if save_path is not None:
            torch.save(self.model.state_dict(), save_path)
            mlflow.log_artifact(save_path)
            mlflow.pytorch.log_model(self.model, "model")
            logging.info(f"Model saved to {save_path} and logged to MLflow")
    
    @staticmethod
    def from_config(config: Dict, dataset: AugmentedDataset):
        trainer_cfg = config.get("trainer", {})
        mlflow_cfg = trainer_cfg.get("mlflow", {})
        env = DTCCTrainer._setup_model_environment(config, dataset)

        return MlFlowDTCCTrainer(
            model=env["model"],
            dataloader=env["dataloader"],
            augment_time_series=dataset.augmentation,
            optimizer=env["optimizer"],
            lambda_cd=trainer_cfg.get("lambda_cd", 1.0),
            num_epochs=trainer_cfg.get("num_epochs", 100),
            update_interval=trainer_cfg.get("update_interval", 5),
            device=env["device"],
            experiment_name=mlflow_cfg.get("experiment", "MLflow_DTCC_Training"),
            run_name=mlflow_cfg.get("run", "default_run")
        )


import torch
from torch.utils.data import Dataset
from abc import abstractmethod

class AugmentedDataset(Dataset):
    def __init__(self, dataframe, feature_cols, target_col):
        self.X = dataframe[feature_cols].values.astype('float32')
        self.y = dataframe[target_col].astype('int64').values

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        x = torch.tensor(self.X[idx])
        if x.ndim == 1:
            x = x.unsqueeze(-1)  # add feature dimension if only batch and seq_len provided
        y = torch.tensor(self.y[idx])
        return x, y
    
    @abstractmethod
    def augmentation(self, batch_x):
        pass


import logging
import torch
import torch.nn as nn
from typing import List
from .helper import EPS, stablize

class DilatedRNN(nn.Module):
    def __init__(self, input_dim, num_layers, hidden_dim, dilation_rates, bidirectional=True):
        super(DilatedRNN, self).__init__()
        self.num_layers = num_layers
        self.hidden_dims = hidden_dim
        self.dilation_rates = dilation_rates
        self.bidirectional = bidirectional
        
        self.rnn_layers = nn.ModuleList()
        for i in range(num_layers):
            if i == 0:
                self.rnn_layers.append(nn.GRU(input_dim, self.hidden_dims[i], batch_first=True, 
                                             bidirectional=bidirectional))
            else:
                input_size = self.hidden_dims[i-1] * (2 if bidirectional else 1)
                self.rnn_layers.append(nn.GRU(input_size, self.hidden_dims[i], batch_first=True, 
                                             bidirectional=bidirectional))
                
    def forward(self, x):
        assert x.ndim == 3, f"Input must be 3D, got {x.shape}"        
        outputs = []
        current_input = x
        
        for i, layer in enumerate(self.rnn_layers):
            # Apply dilation
            dilated_input = self._apply_dilation(current_input, self.dilation_rates[i])
            output, hidden = layer(dilated_input)
            outputs.append(hidden.view(hidden.size(1), -1))  # Extract last hidden state
            current_input = output
            
        # Concatenate the last hidden state from each layer
        return torch.cat(outputs, dim=1)
    
    def _apply_dilation(self, x, dilation_rate):
        """
        Apply dilation to input sequence.
        Args:
            x: [batch_size, seq_len, feature_dim]
            dilation_rate: int
        Returns:
            Dilated x: [batch_size, new_seq_len, feature_dim]
        """
        batch_size, seq_len, feature_dim = x.size()
        if seq_len < dilation_rate:
            # Fallback: just return the last time step (or pad)
            last_step = x[:, -1:, :]  # shape: [batch, 1, feature_dim]
            return last_step
        else:
            # Usual dilation
            indices = torch.arange(0, seq_len, dilation_rate, device=x.device)
            dilated_x = torch.index_select(x, 1, indices)
            return dilated_x


class Encoder(nn.Module):
    def __init__(self, input_dim, num_layers, hidden_dim, dilation_rates, latent_dim):
        super(Encoder, self).__init__()
        self.dilated_rnn = DilatedRNN(input_dim, num_layers, hidden_dim, dilation_rates)
        self.norm = nn.LayerNorm(latent_dim)
        
    def forward(self, x):
        assert x.ndim == 3, f"Input must be 3D, got {x.shape}"
        z = self.dilated_rnn(x)
        return self.norm(z)


class Decoder(nn.Module):
    def __init__(self, latent_dim, output_dim):
        super(Decoder, self).__init__()
        self.rnn = nn.GRU(latent_dim, latent_dim, batch_first=True)
        self.output_layer = nn.Linear(latent_dim, output_dim)
        
    def forward(self, z, seq_len):
        # Expand z to sequence length
        z_expanded = z.unsqueeze(1).repeat(1, seq_len, 1)
        output, _ = self.rnn(z_expanded)
        return self.output_layer(output)

class DTCC(nn.Module):
    def __init__(self, 
                 input_dim: int, 
                 num_layers: int, 
                 num_clusters: int,
                 hidden_dims: List[int], 
                 dilation_rates: List[int], 
                 tau_I: float = 0.5,
                 tau_C: float = 0.5,
                 stable_svd: bool = False):
        super(DTCC, self).__init__()
        latent_dim = sum(hidden_dims) * 2 # the hidden dims times 2 for being bidirectional
        self.encoder = Encoder(input_dim, num_layers, hidden_dims, dilation_rates, latent_dim)
        self.decoder = Decoder(latent_dim, input_dim)
        self.aug_encoder = Encoder(input_dim, num_layers, hidden_dims, dilation_rates, latent_dim)
        self.aug_decoder = Decoder(latent_dim, input_dim)
        self.num_clusters = num_clusters
        self.tau_I = tau_I  # Instance temperature
        self.tau_C = tau_C  # Cluster temperature
        self.stable_svd = stable_svd
        self._init_weights()
    
    def _init_weights(self):
        # Xavier/Glorot init
        for m in self.modules():
            if isinstance(m, nn.GRU):
                for name, param in m.named_parameters():
                    if 'weight' in name:
                        nn.init.xavier_uniform_(param)
                    elif 'bias' in name:
                        nn.init.zeros_(param)
            elif isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, nn.LayerNorm):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
    
    def get_num_clusters(self):
        return self.num_clusters

    def get_stable_svd(self):
        return self.stable_svd

    def forward(self, x, x_aug):
        assert x.ndim == 3, f"Input must be 3D, got {x.shape}"
        # Original view
        z = self.encoder(x)
        x_recon = self.decoder(z, x.size(1))
        
        # Augmented view
        z_aug = self.aug_encoder(x_aug)
        x_aug_recon = self.aug_decoder(z_aug, x_aug.size(1))
        
        return z, z_aug, x_recon, x_aug_recon
    
    def compute_reconstruction_loss(self, x, x_recon, x_aug, x_aug_recon):
        recon_org = torch.mean((x - x_recon) ** 2)
        recon_aug = torch.mean((x_aug - x_aug_recon) ** 2)
        return recon_org + recon_aug
    
    def compute_instance_contrastive_loss(self, z, z_aug):
        n = z.size(0)
        z_norm = nn.functional.normalize(z, dim=1)
        z_aug_norm = nn.functional.normalize(z_aug, dim=1)
        
        # z -> z_aug
        sim_matrix = torch.matmul(z_norm, z_aug_norm.T) / self.tau_I
        max_sim_mat = sim_matrix.max(dim=1, keepdim=True)[0]
        sim_matrix = sim_matrix - max_sim_mat  # shift for stability

        positives = torch.diag(sim_matrix)
        exp_sim = torch.exp(sim_matrix)
        mask = torch.eye(n, device=z.device)
        neg_mask = 1 - mask
        
        # denominator: all except the positive
        denominator = torch.sum(exp_sim * neg_mask, dim=1) + torch.exp(positives)
        # Correct: log(exp(x)/y) = x - log(y)
        instance_loss = -torch.mean(positives - torch.log(denominator + EPS))
        
        # z_aug -> z
        sim_matrix_aug = torch.matmul(z_aug_norm, z_norm.T) / self.tau_I
        max_sim_mat_aug = sim_matrix_aug.max(dim=1, keepdim=True)[0]
        sim_matrix_aug = sim_matrix_aug - max_sim_mat_aug

        positives_aug = torch.diag(sim_matrix_aug)
        exp_sim_aug = torch.exp(sim_matrix_aug)
        denominator_aug = torch.sum(exp_sim_aug * neg_mask, dim=1) + torch.exp(positives_aug)
        instance_loss_aug = -torch.mean(positives_aug - torch.log(denominator_aug + EPS))
        
        return 0.5 * (instance_loss + instance_loss_aug) 
    
    def compute_cluster_distribution_loss(self, z, z_aug):
        assert not torch.isnan(z).any(), "NaN in z before SVD"
        assert not torch.isinf(z).any(), "Inf in z before SVD"
        assert not torch.isnan(z_aug).any(), "NaN in z_aug before SVD"
        assert not torch.isinf(z_aug).any(), "Inf in z_aug before SVD"
        
        logging.debug(f"original z:\n {z}")
        logging.debug(f"original z_aug:\n {z_aug}")
        if self.stable_svd:
            z = stablize(z)
            logging.debug(f"stablized z:\n {z}")
            z_aug = stablize(z_aug)
            logging.debug(f"stablized z_aug:\n {z_aug}")

        # SVD for original view
        U, S, V = torch.linalg.svd(z)  # z: [batch_size, latent_dim]
        Q = U[:, :self.num_clusters]  # [batch_size, num_clusters]
        assert not torch.isnan(Q).any(), "NaN in Q after SVD"

        # SVD for augmented view
        
        U_aug, S_aug, V_aug = torch.linalg.svd(z_aug)
        Q_aug = U_aug[:, :self.num_clusters]
        assert not torch.isnan(Q_aug).any(), "NaN in Q_aug after SVD"

        # Compute k-means loss
        gram_matrix = torch.matmul(z.T, z)           # [latent_dim, latent_dim]
        gram_matrix_aug = torch.matmul(z_aug.T, z_aug)  # [latent_dim, latent_dim]

        # Compute QTZQ = (z.T @ Q).T @ (z.T @ Q)
        ZQ = torch.matmul(z.T, Q)        # [latent_dim, num_clusters]
        QTZQ = torch.matmul(ZQ.T, ZQ)    # [num_clusters, latent_dim] @ [latent_dim, num_clusters] = [num_clusters, num_clusters]

        ZQ_aug = torch.matmul(z_aug.T, Q_aug)
        QTZQ_aug = torch.matmul(ZQ_aug.T, ZQ_aug)

        km_org = torch.trace(gram_matrix) - torch.trace(QTZQ)
        km_aug = torch.trace(gram_matrix_aug) - torch.trace(QTZQ_aug)

        return 0.5 * (km_org + km_aug), Q, Q_aug
    
    def compute_cluster_contrastive_loss(self, Q, Q_aug):
        logging.debug("Q:\n{}".format(Q))
        logging.debug("Q_aug:\n{}".format(Q_aug))
        k = Q.size(1)
        
        # Normalize Q and Q_aug
        Q_norm = nn.functional.normalize(Q, dim=0)
        Q_aug_norm = nn.functional.normalize(Q_aug, dim=0)
        
        # Compute similarity matrix
        sim_matrix = torch.matmul(Q_norm.T, Q_aug_norm) / self.tau_C
        
        # Positive pairs are on the diagonal
        positives = torch.diag(sim_matrix)
        
        # For each q_i, compute denominator
        exp_sim = torch.exp(sim_matrix)
        mask = torch.eye(k, device=Q.device)
        neg_mask = 1 - mask
        
        denominator = torch.sum(exp_sim * neg_mask, dim=1) + torch.exp(positives)
        cluster_loss = -torch.mean(torch.log(torch.exp(positives) / (denominator+EPS) + EPS))
        
        # Compute entropy term
        P_q = torch.mean(Q, dim=0)
        # tends to go down to negative close zero
        P_q_aug = torch.mean(Q_aug, dim=0)
        # Apply clamp for numerical stability
        P_q = torch.clamp(P_q, min=EPS)
        P_q_aug = torch.clamp(P_q_aug, min=EPS)
        logging.debug("P_q:\n{}".format(P_q))
        logging.debug("P_q_aug:\n{}".format(P_q_aug))
        entropy = -torch.sum(P_q * torch.log(P_q)) - torch.sum(P_q_aug * torch.log(P_q_aug))
        
        return cluster_loss - entropy


import numpy as np
import torch
from typing import Optional, Callable

def check_shape(x):
    assert x.ndim == 3, f"Input must be [batch, seq_len, features], got {x.shape}"

def ensure_float32(x):
    if x.dtype != np.float32:
        return x.astype(np.float32)
    return x

def torch_augmentation_wrapper(
    aug_fn: Callable, 
    x: torch.Tensor, 
    labels: Optional[torch.Tensor] = None,
    **kwargs
) -> torch.Tensor:
    """
    Wrapper to apply numpy augmentations on torch tensors.
    """
    x_np = x.detach().cpu().numpy()
    if labels is not None:
        labels_np = labels.detach().cpu().numpy()
        aug_np = aug_fn(x_np, labels_np, **kwargs)
    else:
        aug_np = aug_fn(x_np, **kwargs)
    return torch.from_numpy(aug_np).to(x.device).type_as(x)

# def random_time_series_augmentation(
#     x: np.ndarray,
#     labels: Optional[np.ndarray] = None,
#     random_state: Optional[int] = None,
#     allow_label_dependent: bool = False
# ) -> np.ndarray:
#     """
#     Randomly applies a time-series augmentation.
#     If labels are provided and allow_label_dependent=True, may select label-dependent augmentations.
#     """
#     _check_shape(x)
#     rng = np.random.RandomState(random_state)
#     # List of (func, needs_labels)
#     aug_funcs: List[Tuple[Callable, bool]] = [
#         (jitter, False),
#         (scaling, False),
#         (rotation, False),
#         (permutation, False),
#         (magnitude_warp, False),
#         (time_warp, False),
#         (window_slice, False),
#         (window_warp, False),
#     ]
#     if allow_label_dependent and labels is not None:
#         # These require labels and external DTW code, so only include if possible
#         try:
#             from utils import dtw  # Will raise ImportError if not available
#             aug_funcs.extend([
#                 # (spawner, True),  # Uncomment if dtw is available and tested
#                 # (wdba, True),
#                 # (random_guided_warp, True),
#                 # (random_guided_warp_shape, True),
#                 # (discriminative_guided_warp, True),
#                 # (discriminative_guided_warp_shape, True),
#             ])
#         except ImportError:
#             pass

#     idx = rng.randint(0, len(aug_funcs))
#     func, needs_labels = aug_funcs[idx]
#     if needs_labels and labels is not None:
#         return func(x, labels, random_state=random_state)
#     return func(x, random_state=random_state)



import numpy as np
from scipy.special import comb
from sklearn.metrics import normalized_mutual_info_score, adjusted_rand_score
from scipy.optimize import linear_sum_assignment

def cluster_acc(y_true, y_pred):
    y_true = np.asarray(y_true).astype(np.int64)
    y_pred = np.asarray(y_pred).astype(np.int64)
    assert y_pred.size == y_true.size
    D = max(y_pred.max(), y_true.max()) + 1
    w = np.zeros((D, D), dtype=np.int64)
    for i in range(y_pred.size):
        w[y_pred[i], y_true[i]] += 1
    row_ind, col_ind = linear_sum_assignment(w.max() - w)
    total = 0
    for i, j in zip(row_ind, col_ind):
        total += w[i, j]
    return total * 1.0 / y_pred.size

def nmi_score(y_true, y_pred):
    return normalized_mutual_info_score(y_true, y_pred)

def ari_score(y_true, y_pred):
    return adjusted_rand_score(y_true, y_pred)

def rand_index_score(clusters, classes):
    tp_plus_fp = comb(np.bincount(clusters), 2).sum()
    tp_plus_fn = comb(np.bincount(classes), 2).sum()
    A = np.c_[(clusters, classes)]
    tp = sum(comb(np.bincount(A[A[:, 0] == i, 1]), 2).sum()
             for i in set(clusters))
    fp = tp_plus_fp - tp
    fn = tp_plus_fn - tp
    tn = comb(len(A), 2) - tp - fp - fn
    return (tp + tn) / (tp + fp + fn + tn)

def all_metrics(y_true, y_pred):
    return {
        "acc": cluster_acc(y_true, y_pred),
        "nmi": nmi_score(y_true, y_pred),
        "ari": ari_score(y_true, y_pred),
        "ri": rand_index_score(y_pred, y_true)
    }


from scipy.io import arff
import pandas as pd
from torchdtcc.augmentations.basic import jitter
from torchdtcc.augmentations.helper import torch_augmentation_wrapper
from ..augmented_dataset import AugmentedDataset

class MeatArffDataset(AugmentedDataset):
    def __init__(self, path):
        data_train, _ = arff.loadarff(path + 'Meat_TRAIN.arff')
        data_test, _ = arff.loadarff(path + 'Meat_TEST.arff')
        df_train = pd.DataFrame(data_train)
        df_test = pd.DataFrame(data_test)
        df = pd.concat([df_train, df_test], ignore_index=True)
        for col in df.select_dtypes([object]):
            df[col] = df[col].apply(lambda x: x.decode('utf-8') if isinstance(x, bytes) else x)
        feature_cols = [col for col in df.columns if col.startswith('att')]
        target_col = 'target'
        # Ensure target is int
        df[target_col] = df[target_col].astype(int)
        super().__init__(df, feature_cols, target_col)

    def augmentation(self, batch_x):        
        # Ensure [batch, seq_len, features]
        assert batch_x.ndim == 3, f"Input must be 3D, got {batch_x.shape}"
        
        x_aug = torch_augmentation_wrapper(jitter, batch_x)
        # x_augx = torch_augmentation_wrapper(scaling, x_aug) # Import scaling first

        return x_aug



import torch
import torch.optim as optim
from torch.utils.data import DataLoader
from tqdm import tqdm
import logging
from typing import Dict
from ..dtcc.clustering import Clusterer
from ..dtcc.dtcc import DTCC
from torchdtcc.datasets.augmented_dataset import AugmentedDataset

class DTCCTrainer:
    def __init__(
        self, 
        model: DTCC, 
        dataloader: DataLoader, 
        augment_time_series, 
        optimizer, 
        lambda_cd, 
        num_epochs, 
        update_interval=5, 
        device="cpu"
    ):
        self.model = model
        self.device = device
        self.clusterer = Clusterer(self.device)
        self.dataloader = dataloader
        self.augment_time_series = augment_time_series
        self.optimizer = optimizer
        self.lambda_cd = lambda_cd
        self.num_epochs = num_epochs
        self.update_interval = update_interval

    def run(self, save_path=None):
        self.model.train()
        for epoch in range(self.num_epochs):
            epoch_loss = 0
            recon_losses, instance_losses, cd_losses, cluster_losses, total_losses = [], [], [], [], []
            with tqdm(total=len(self.dataloader), desc=f'Epoch {epoch+1}/{self.num_epochs}') as pbar:
                for i, batch in enumerate(self.dataloader):
                    x, y = batch
                    x = x.to(self.device)
                    x_aug = self.augment_time_series(x)
                    z, z_aug, x_recon, x_aug_recon = self.model(x, x_aug)
                    recon_loss = self.model.compute_reconstruction_loss(x, x_recon, x_aug, x_aug_recon)
                    instance_loss = self.model.compute_instance_contrastive_loss(z, z_aug)
                    cd_loss, Q, Q_aug = self.model.compute_cluster_distribution_loss(z, z_aug)
                    cluster_loss = self.model.compute_cluster_contrastive_loss(Q, Q_aug)
                    loss = recon_loss + instance_loss + cluster_loss + self.lambda_cd * cd_loss

                    recon_losses.append(recon_loss.item())
                    instance_losses.append(instance_loss.item())
                    cd_losses.append(cd_loss.item())
                    cluster_losses.append(cluster_loss.item())
                    total_losses.append(loss.item())

                    logging.debug(f"Step {i} | recon: {recon_loss.item():.4f} | instance: {instance_loss.item():.4f} | cd: {cd_loss.item():.4f} | cluster: {cluster_loss.item():.4f} | total: {loss.item():.4f}")

                    self.optimizer.zero_grad()
                    loss.backward()
                    self.optimizer.step()
                    epoch_loss += loss.item()

                    pbar.set_postfix({'recon': f"{recon_loss.item():.3f}", 'inst': f"{instance_loss.item():.3f}", 'cd': f"{cd_loss.item():.3f}", 'clust': f"{cluster_loss.item():.3f}", 'total': f"{loss.item():.3f}"})
                    pbar.update(1)

            avg_recon = sum(recon_losses) / len(recon_losses)
            avg_instance = sum(instance_losses) / len(instance_losses)
            avg_cd = sum(cd_losses) / len(cd_losses)
            avg_cluster = sum(cluster_losses) / len(cluster_losses)
            avg_total = sum(total_losses) / len(total_losses)

            self.log_loss(epoch, avg_recon, avg_instance, avg_cd, avg_cluster, avg_total)

            if (epoch + 1) % self.update_interval == 0:
                self.clusterer.set_model(self.model)
                metrics = self.clusterer.evaluate(self.dataloader)
                self.log_evaluation(epoch, metrics)  # Fixed method name

        self.save_model(save_path)  # Moved outside loop to save only final model
        return self.model

    def log_loss(self, epoch, avg_recon, avg_instance, avg_cd, avg_cluster, avg_total):
        logging.info(f"Epoch {epoch+1}/{self.num_epochs} | avg recon: {avg_recon:.4f} | avg instance: {avg_instance:.4f} | avg cd: {avg_cd:.4f} | avg cluster: {avg_cluster:.4f} | avg total: {avg_total:.4f}")

    def log_evaluation(self, epoch, metrics):  # Fixed method name
        print(f"Epoch {epoch+1}: ACC={metrics['acc']:.4f} NMI={metrics['nmi']:.4f} ARI={metrics['ari']:.4f} RI={metrics['ri']:.4f}")

    def save_model(self, save_path):
        if save_path is not None:
            torch.save(self.model.state_dict(), save_path)
            logging.info(f"Model saved to {save_path}")

    @staticmethod
    def _setup_model_environment(config: Dict, dataset: AugmentedDataset):
        model_cfg = config["model"]
        data_cfg = config["data"]
        trainer_cfg = config.get("trainer", {})
        device = torch.device(config["device"] if torch.cuda.is_available() else "cpu")

        dataloader = DataLoader(dataset, batch_size=data_cfg["batch_size"], shuffle=True)

        model = DTCC(
            input_dim=model_cfg["input_dim"],
            num_layers=model_cfg["num_layers"],
            num_clusters=model_cfg["num_clusters"],
            hidden_dims=model_cfg["hidden_dims"],
            dilation_rates=model_cfg["dilation_rates"],
            tau_I=model_cfg["tau_I"],
            tau_C=model_cfg["tau_C"],
            stable_svd=model_cfg["stable_svd"]
        ).to(device)

        optimizer = optim.Adam(
            model.parameters(),
            lr=trainer_cfg.get("learning_rate", 1e-3),
            weight_decay=trainer_cfg.get("weight_decay", 0)
        )
        return {
            "dataloader": dataloader,
            "model": model,
            "optimizer": optimizer,
            "device": device
        }

    @staticmethod
    def from_config(config: Dict, dataset: AugmentedDataset):
        trainer_cfg = config.get("trainer", {})
        env = DTCCTrainer.setup_model_environment(config, dataset)

        return DTCCTrainer(
            model=env["model"],
            dataloader=env["dataloader"],
            augment_time_series=dataset.augmentation,
            optimizer=env["optimizer"],
            lambda_cd=trainer_cfg.get("lambda_cd", 1.0),
            num_epochs=trainer_cfg.get("num_epochs", 100),
            update_interval=trainer_cfg.get("update_interval", 5),
            device=env["device"]
        )


import itertools
from copy import deepcopy
import torch
from typing import Dict, List, Tuple, Callable
import logging
from .trainer import DTCCTrainer
from torchdtcc.dtcc.clustering import Clusterer
from torchdtcc.datasets.augmented_dataset import AugmentedDataset

class HyperparameterSearch:
    def __init__(
            self, 
            base_config: Dict, 
            dataset: AugmentedDataset, 
            tau_I_values: List[float], 
            tau_C_values: List[float],
            trainer_factory: Callable[[Dict, AugmentedDataset], DTCCTrainer]
        ):
        self.base_config = base_config
        self.dataset = dataset
        self.tau_I_values = tau_I_values
        self.tau_C_values = tau_C_values
        self.device = torch.device(base_config["device"] if torch.cuda.is_available() else "cpu")
        self.results = []
        self.trainer_factory = trainer_factory

    def grid_search(self) -> List[Dict]:
        param_combinations = list(itertools.product(self.tau_I_values, self.tau_C_values))
        logging.info(f"Starting grid search over {len(param_combinations)} combinations of tau_I and tau_C")

        for tau_I, tau_C in param_combinations:
            logging.info(f"Testing tau_I={tau_I}, tau_C={tau_C}")
            # Update config with current hyperparameters
            config = deepcopy(self.base_config)
            config["model"]["tau_I"] = tau_I
            config["model"]["tau_C"] = tau_C
            config["trainer"]["mlflow"]["run"] += f"_tau_I_{tau_I}_tau_C_{tau_C}"

            # Initialize trainer
            trainer = self.trainer_factory(config, self.dataset)
            save_path = config.get("trainer", {}).get("save_path", f"dtcc_model_tauI_{tau_I}_tauC_{tau_C}.pth")

            # Run training
            model = trainer.run(save_path=save_path)

            # Evaluate clustering metrics (assuming last update_interval metrics are available)
            clusterer = Clusterer(self.device)
            clusterer.set_model(model)
            metrics = clusterer.evaluate(trainer.dataloader)

            # Store results
            result = {
                "tau_I": tau_I,
                "tau_C": tau_C,
                "metrics": metrics,
                "save_path": save_path
            }
            self.results.append(result)
            logging.info(f"Results for tau_I={tau_I}, tau_C={tau_C}: ACC={metrics['acc']:.4f}, NMI={metrics['nmi']:.4f}")

        return self.results

    def get_best_params(self, metric: str = "acc") -> Dict:
        if not self.results:
            return {}
        return max(self.results, key=lambda x: x["metrics"][metric])


import numpy as np
from .helper import check_shape, ensure_float32

def jitter(x, sigma=0.03, random_state=None):
    """
    Adds Gaussian noise to the input.
    Reference: https://arxiv.org/pdf/1706.00527.pdf
    """
    check_shape(x)
    rng = np.random.RandomState(random_state)
    return ensure_float32(x + rng.normal(loc=0., scale=sigma, size=x.shape))

def scaling(x, sigma=0.1, random_state=None):
    """
    Multiplies input by a random scaling factor for each feature.
    Reference: https://arxiv.org/pdf/1706.00527.pdf
    """
    check_shape(x)
    rng = np.random.RandomState(random_state)
    factor = rng.normal(loc=1., scale=sigma, size=(x.shape[0], x.shape[2]))
    return ensure_float32(np.multiply(x, factor[:, np.newaxis, :]))

def rotation(x, random_state=None):
    """
    Randomly flips and permutes feature axes.
    """
    check_shape(x)
    rng = np.random.RandomState(random_state)
    flip = rng.choice([-1, 1], size=(x.shape[0], x.shape[2]))
    rotate_axis = np.arange(x.shape[2])
    rng.shuffle(rotate_axis)
    return ensure_float32(flip[:, np.newaxis, :] * x[:, :, rotate_axis])

def permutation(x, max_segments=5, seg_mode="equal", random_state=None):
    """
    Randomly permutes segments of the time series.
    """
    check_shape(x)
    rng = np.random.RandomState(random_state)
    orig_steps = np.arange(x.shape[1])
    num_segs = rng.randint(1, max_segments, size=(x.shape[0]))
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        if num_segs[i] > 1:
            if seg_mode == "random":
                split_points = rng.choice(x.shape[1]-2, num_segs[i]-1, replace=False)
                split_points.sort()
                splits = np.split(orig_steps, split_points)
            else:
                splits = np.array_split(orig_steps, num_segs[i])
            warp = np.concatenate(rng.permutation(splits)).ravel()
            ret[i] = pat[warp]
        else:
            ret[i] = pat
    return ensure_float32(ret)

def magnitude_warp(x, sigma=0.2, knot=4, random_state=None):
    """
    Warps the magnitude of the series using random smooth curves.
    """
    from scipy.interpolate import CubicSpline
    check_shape(x)
    rng = np.random.RandomState(random_state)
    orig_steps = np.arange(x.shape[1])
    random_warps = rng.normal(loc=1.0, scale=sigma, size=(x.shape[0], knot+2, x.shape[2]))
    warp_steps = (np.ones((x.shape[2],1))*(np.linspace(0, x.shape[1]-1., num=knot+2))).T
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        warper = np.array([CubicSpline(warp_steps[:,dim], random_warps[i,:,dim])(orig_steps) for dim in range(x.shape[2])]).T
        ret[i] = pat * warper
    return ensure_float32(ret)

def time_warp(x, sigma=0.2, knot=4, random_state=None):
    """
    Warps the time axis using smooth curves.
    """
    from scipy.interpolate import CubicSpline
    check_shape(x)
    rng = np.random.RandomState(random_state)
    orig_steps = np.arange(x.shape[1])
    random_warps = rng.normal(loc=1.0, scale=sigma, size=(x.shape[0], knot+2, x.shape[2]))
    warp_steps = (np.ones((x.shape[2],1))*(np.linspace(0, x.shape[1]-1., num=knot+2))).T
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        for dim in range(x.shape[2]):
            time_warp_curve = CubicSpline(warp_steps[:,dim], warp_steps[:,dim] * random_warps[i,:,dim])(orig_steps)
            scale = (x.shape[1]-1)/time_warp_curve[-1] if time_warp_curve[-1] != 0 else 1.0
            ret[i,:,dim] = np.interp(orig_steps, np.clip(scale*time_warp_curve, 0, x.shape[1]-1), pat[:,dim]).T
    return ensure_float32(ret)

def window_slice(x, reduce_ratio=0.9, random_state=None):
    """
    Randomly slices a window from the time series and rescales back to original size.
    Reference: https://halshs.archives-ouvertes.fr/halshs-01357973/document
    """
    check_shape(x)
    rng = np.random.RandomState(random_state)
    target_len = np.ceil(reduce_ratio*x.shape[1]).astype(int)
    if target_len >= x.shape[1]:
        return ensure_float32(x)
    starts = rng.randint(low=0, high=x.shape[1]-target_len, size=(x.shape[0])).astype(int)
    ends = (target_len + starts).astype(int)
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        for dim in range(x.shape[2]):
            ret[i,:,dim] = np.interp(np.linspace(0, target_len-1, num=x.shape[1]), np.arange(target_len), pat[starts[i]:ends[i],dim]).T
    return ensure_float32(ret)

def window_warp(x, window_ratio=0.1, scales=[0.5, 2.], random_state=None):
    """
    Warps a random window in the series by a random scale and resamples to original length.
    Reference: https://halshs.archives-ouvertes.fr/halshs-01357973/document
    """
    check_shape(x)
    rng = np.random.RandomState(random_state)
    warp_scales = rng.choice(scales, x.shape[0])
    warp_size = np.ceil(window_ratio*x.shape[1]).astype(int)
    window_steps = np.arange(warp_size)
    window_starts = rng.randint(low=1, high=x.shape[1]-warp_size-1, size=(x.shape[0])).astype(int)
    window_ends = (window_starts + warp_size).astype(int)
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        for dim in range(x.shape[2]):
            start_seg = pat[:window_starts[i],dim]
            window_seg = np.interp(np.linspace(0, warp_size-1, num=int(warp_size*warp_scales[i])), window_steps, pat[window_starts[i]:window_ends[i],dim])
            end_seg = pat[window_ends[i]:,dim]
            warped = np.concatenate((start_seg, window_seg, end_seg))
            ret[i,:,dim] = np.interp(np.arange(x.shape[1]), np.linspace(0, warped.size-1, num=x.shape[1]), warped).T
    return ensure_float32(ret)


import torch
import numpy as np
import logging
from sklearn.cluster import KMeans
from .dtcc import DTCC
from .helper import stablize
from ..training.metrics import all_metrics

class Clusterer:
    def __init__(self, device):
        self.device = torch.device(device if torch.cuda.is_available() else "cpu")
        pass

    def set_model(self, model: DTCC):
        """
        Args:
            model (DTCC): trained DTCC model object
            num_clusters (int): number of clusters the model is able to cluster
        """
        self.model = model
        self.num_clusters = model.get_num_clusters()
        self.stable_svd = model.get_stable_svd()

    def load_model(self, model_path, model_kwargs, device="cpu"):
        """
        Args:
            model_path (str): Path to model weights (.pth)
            model_class (type): Model class (e.g., DTCC)
            model_kwargs (dict): Model init kwargs
            device (str): 'cpu' or 'cuda'
        """
        self.device = torch.device(device if torch.cuda.is_available() else "cpu")
        self.model = DTCC(**model_kwargs)
        self.model.load_state_dict(torch.load(model_path, map_location=self.device))
        self.model.to(self.device)
        self.model.eval()
        self.num_clusters = model_kwargs.get("num_clusters", 3)
        self.stable_svd = model_kwargs.get("stable_svd", False)

    def encode(self, dataloader):
        zs = []
        with torch.no_grad():
            for batch in dataloader:
                X, y = batch
                X = X.to(self.device)
                z = self.model.encoder(X)
                zs.append(z)
        return torch.cat(zs, dim=0)  # shape [N, d]

    def soft_clusters(self, z_all):
        if self.stable_svd:
            z_all = stablize(z_all)
        U, S, V = torch.linalg.svd(z_all)
        Q = U[:, :self.num_clusters]
        return Q

    def hard_clusters_argmax(self, Q):
        return Q.argmax(dim=1)

    def hard_clusters_kmeans(self, Q):
        Q_np = Q.cpu().numpy()
        kmeans = KMeans(n_clusters=self.num_clusters, n_init=10)
        return kmeans.fit_predict(Q_np)

    def cluster(self, dataloader, method="kmeans"):
        z_all = self.encode(dataloader)
        Q = self.soft_clusters(z_all)
        if method == "soft":
            return Q
        elif method == "argmax":
            return self.hard_clusters_argmax(Q)
        elif method == "kmeans":
            return self.hard_clusters_kmeans(Q)
        else:
            raise ValueError(f"Unknown clustering method: {method}")
        
    def evaluate(self, dataloader, metrics="all", method="kmeans"):
        predicted = self.cluster(dataloader=dataloader, method=method)

        ground_truth = []
        with torch.no_grad():
            for batch in dataloader:
                X, y = batch
                ground_truth.append(y)
        ground_truth = torch.cat(ground_truth, dim=0)

        # Ensure both are 1D numpy arrays
        ground_truth = ground_truth.cpu().numpy().reshape(-1)
        if isinstance(predicted, torch.Tensor):
            predicted = predicted.cpu().numpy().reshape(-1)
        else:
            predicted = np.array(predicted).reshape(-1)
        logging.debug(f"Shapes: {ground_truth.shape} vs {predicted.shape}")
        assert ground_truth.shape == predicted.shape, f"Shape mismatch: {ground_truth.shape} vs {predicted.shape}"

        return all_metrics(ground_truth, predicted)